// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This is bigsea management driver for Linux.
 * Bigsea is a video encoder, has two cores, only support vp9 format.
 * This driver provide some IOCTL commands for userspace,
 * like reserve/release a hardware core, access registers.
 * At the same time, bigsea and vc8000e only one can work, so they use
 * same functions to reserve/release core.
 * How to operate Bigsea: reserve a idle core, config registers,
 * enable the core, wait interrupt, release the core.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <time.h>

#include "hugetlbfs.h"
#include "transcoder.h"

#define TRANS_SIZE 10 //MB
#define TRANS_EP_S0_START_ADDR 0x11800000UL
#define TRANS_EP_S0_END_ADDR 0xE0000000UL
#define TRANS_EP_S1_START_ADDR 0x200000000UL
#define TRANS_EP_S1_END_ADDR 0x2E0000000UL

struct t_info {
	char *rc_src;
	char *rc_back;
	int fd_dev;
	int size;
	char slice;
};

static int do_trans_test(struct t_info *ti)
{
	unsigned long long i;
	int fd = ti->fd_dev;
	int size = ti->size;
	unsigned long long ep_start_addr;
	unsigned long long ep_end_addr;
	int count = 0;
	struct trans_pcie_hdma edma_trans_wr, edma_trans_rd;

	ti->rc_src = get_huge_pages(size, GHP_DEFAULT);
	if (ti->rc_src == NULL) {
		printf("get_huge_pages failed\n");
		return -1;
	}
	ti->rc_back = get_huge_pages(size, GHP_DEFAULT);
	if (ti->rc_back == NULL) {
		printf("get_huge_pages failed\n");
		return -1;
	}

	// fill random data
	srand((unsigned int)time(NULL));
	for (i = 0; i < size; i++) {
		ti->rc_src[i] = rand() % 256;
	}

	if (ti->slice == 0) {
		ep_start_addr = TRANS_EP_S0_START_ADDR;
		ep_end_addr = TRANS_EP_S0_END_ADDR;
	} else {
		ep_start_addr = TRANS_EP_S1_START_ADDR;
		ep_end_addr = TRANS_EP_S1_END_ADDR;
	}

	do {
		for (i = ep_start_addr; i <= (ep_end_addr - size); i += size) {
			edma_trans_wr.dir = RC2EP;
			edma_trans_wr.size = size;
			edma_trans_wr.sar = (unsigned long long)ti->rc_src;
			edma_trans_wr.dar = i;
			if (ioctl(fd, SN_TRANX_HDMA_TRANX_VIR, &edma_trans_wr) <
			    0) {
				printf("write to 0x%llx failed!\n", i);
				return -1;
			}

			edma_trans_rd.dir = EP2RC;
			edma_trans_rd.size = size;
			edma_trans_rd.sar = i;
			edma_trans_rd.dar = (unsigned long long)ti->rc_back;
			if (ioctl(fd, SN_TRANX_HDMA_TRANX_VIR, &edma_trans_rd) <
			    0) {
				printf("read from 0x%llx failed!\n", i);
				return -1;
			}
			if (memcmp(ti->rc_back, ti->rc_src, size) == 0) {
				printf("Verify 0x%llx success, total count=%lld MB!\n",
				       i, (i - ep_start_addr) / 1024 / 1024);
			} else {
				printf("Verify 0x%llx failed!\n", i);
				return -1;
			}
		}
	} while (0);

	return 0;
}

int main(int argc, char **argv)
{
	int ret;
	struct t_info trans_info;
	int fd = -1;
    int mem_size = TRANS_SIZE;

	if (argc < 2) {
		printf("S0/S1 DDR write test. write data to DDR and read back to verify.\n");
		printf("command line: %s <node> <size in MB (%d MB)>.\n", argv[0], TRANS_SIZE);
		printf("example: %s /dev/ama_transcoder0 100\n", argv[0]);
		return -1;
	}

	if (argc < 3) {
		printf("use default mem_size=%dMB\n", mem_size);
	} else
		mem_size = strtoul(argv[2], 0, 0);

	trans_info.size = mem_size*1024*1024;
	trans_info.fd_dev = open(argv[1], O_RDWR, S_IRUSR | S_IWUSR);
	if (trans_info.fd_dev == -1) {
		printf("failed to open %s!\n", argv[1]);
		return -1;
	}

	trans_info.slice = 0;
	ret = do_trans_test(&trans_info);
	printf("slice0 test %s!\n", ret ? "failed" : "success");

	trans_info.slice = 1;
	ret = do_trans_test(&trans_info);
	printf("slice1 test %s!\n", ret ? "failed" : "success");

exit:

	if (trans_info.rc_src)
		free_huge_pages(trans_info.rc_src);
	if (trans_info.rc_back)
		free_huge_pages(trans_info.rc_back);

	close(trans_info.fd_dev);
	return ret;
}
