
// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2020 Xilinx Inc.
 *
 * This is xilinx scaler driver for Linux.
 * This file provide register operation and initialization,
 * like read/write a register or pull/push a batch of registers.
 */

// This is being used as a *reference* implementation should events be needed elsewhere
// They are *not* needed here but used as a PoC for Windows reference

// #define USE_EVENTS

#include <linux/pci.h>
#include <linux/pagemap.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/kfifo.h>

#if defined(USE_EVENTS)
#include <linux/fdtable.h>
#include <linux/eventfd.h>
#endif

#include "common.h"
#include "transcoder.h"
#include "hw_monitor.h"
#include "regs.h"

#include "sn_perf.h"
#include "sn_osal.h"
#include "memory_osal.h"

#if defined(TRACE_XAV1_ENC)
#define TRACE(...) trace_printk(__VA_ARGS__); // msleep(1)
#else
#define TRACE(...)
#endif

#define XAV1_ENC_REGMAP_SIZE            (5*1024*1024) //5MB

#define XAV1_HOST_CONTROL_BASE          (0x400000) //(0x12900000 - 0x12500000)
#define XAV1_HOST_CONTROL_AP            (XAV1_HOST_CONTROL_BASE + 0x00)
#define XAV1_HOST_CONTROL_GIE           (XAV1_HOST_CONTROL_BASE + 0x04)
#define XAV1_HOST_CONTROL_ISR           (XAV1_HOST_CONTROL_BASE + 0x0C)

#define XAV1_MBOX_BASE                  (0x401000) //(0x12902000 - 0x12500000)
#define XAV1_MBOX0_DATA                 (XAV1_MBOX_BASE + 0 * 16)
#define XAV1_MBOX0_LAST                 (XAV1_MBOX0_DATA + 4)
#define XAV1_MBOX0_PTR                  (XAV1_MBOX0_LAST + 4)
#define XAV1_MBOX15_DATA                (XAV1_MBOX_BASE + 15 * 16)
#define XAV1_MBOX15_PTR                 (XAV1_MBOX15_DATA + 8)

#define XAV1_SAR_BASE                   (0x402000) //(0x12902000 - 0x12500000)
#define XAV1_SAR_CONTROL0_SET             (XAV1_SAR_BASE + 0x90)
#define XAV1_SAR_CONTROL0_CLR             (XAV1_SAR_BASE + 0x94)
#define XAV1_SAR_CONTROL0_VAL             (XAV1_SAR_BASE + 0x98)
#define XAV1_SAR_CONTROL1_SET             (XAV1_SAR_BASE + 0xD0)
#define XAV1_SAR_CONTROL1_CLR             (XAV1_SAR_BASE + 0xD4)
#define XAV1_SAR_CONTROL1_VAL             (XAV1_SAR_BASE + 0xD8)
#define XAV1_SAR_HOST_INTR_STATUS       (XAV1_SAR_BASE + 0x0A0)
#define XAV1_SAR_FPS_STATUS             (XAV1_SAR_BASE + 0x0A8)
#define XAV1_SAR_FPS_STATUS1            (XAV1_SAR_BASE + 0x0E0)
#define XAV1_SAR_FPS_VERSION            (XAV1_SAR_BASE + 0x0AC)

#define XAV1_SAR_ECC_COUNTS             (XAV1_SAR_BASE + 0x0B0)

#define XBIT_MASK(bit)                  ((u32)1<<(bit))

struct xav1_enc_core_desc {
    int slice;
    long base;
    int iosize;
    int irq;
    const char *name;
};

/*here config every core in all subsystems*/
//PF Mode
static struct xav1_enc_core_desc pf_core_array[]= {
    {0, S1_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 18, "xav1_enc_0"},
    {1, S2_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 28, "xav1_enc_1"},
};

//ONE VF Mode
static struct xav1_enc_core_desc onevf_core_array[]= {
    {0, ONE_VF_S1_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 18, "xav1_enc_0"},
    {1, ONE_VF_S2_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 28, "xav1_enc_1"},
};

struct core_cfg {
    int irq;
    unsigned long base_addr;
    volatile u8 *hwreg;
    u32 irq_recvd;
    u32 iosize;
    u32 reg_size;
    u32 *shadow;
    const char *name;
    int irq_count;
};

#define MAX_SLOTS_PER_SLICE 128 // DO *NOT* ADJUST THIS
#define MAX_PENDING_MSGS_PER_SLOT 32
#define MAX_WORDS_PER_MSG 4

#define CHANNEL_CONFIG_SIZE_WORDS 96

struct fifo_msg {
    u32 words[MAX_WORDS_PER_MSG];
};

struct slot {
    wait_queue_head_t wait_queue;
    STRUCT_KFIFO(struct fifo_msg, 512) fifo;
    spinlock_t fifo_lock; // TODO(jkuhn): might not need
    struct file *filp;
    int pendingDelete;

    uint32_t task_id; // assumption that all bitstream descriptors / recon come from same task!

    uint8_t* h2fw_ring;
    uint32_t h2fw_write;

    uint8_t* fw2h_ring;
    uint32_t fw2h_write;

    bool fatal_error;

#if defined(USE_EVENTS)
    struct file* eventFile;
#endif
};

struct slice {
    int valid;
    int numChan;
    int nextAvailSlot;
    struct mutex lock;
    struct slot slots[MAX_SLOTS_PER_SLICE];
};

struct xav1_enc {
    struct core_cfg subsys[MAX_SLICE_NUM];
    spinlock_t irq_lock;
    spinlock_t reserve_lock;
    wait_queue_head_t wait_queue;
    wait_queue_head_t reserve_queue;
    void *regs_shadow;
    int isCoreIdle[MAX_SLICE_NUM];
    int isCoreLocked[MAX_SLICE_NUM];
    struct sn_tranx_t *tdev;
    struct slice slices[MAX_SLICE_NUM];
    SnPerfHandle perfHandles[SN_PERF_ID_COUNT];
    u32 mb_full[2];
    struct semaphore boot_sem;
    struct file* boot_sem_owner;
    bool fatal_fw_error;
    struct timer_list ecc_counts_timer;
    uint32_t ecc_counts[2][8][2];
};

struct xav1_slot {
	__u32 slice; // slice in (always)
	__u32 slot; // slice out on alloc, in on free
};

static struct xav1_enc_core_desc* get_core_desc(struct xav1_enc *xav1e, int *num_cores)
{
    struct xav1_enc_core_desc *cp = NULL;
    int ncores = 0;

    switch(xav1e->tdev->pf_vf_mode)
    {
        case PF_MODE:
            if (xav1e->tdev->vf_index != PF_INDEX) {
                sn_pri(xav1e->tdev, SN_ERR, "%s,vf_index incorrect in PF_MODE\n", __func__);
            } else {
                ncores = sizeof(pf_core_array)/sizeof(pf_core_array[0]);
                cp     = pf_core_array;
            }
            break;

        case ONE_VF_MODE:
            if (xav1e->tdev->vf_index != VF1_INDEX) {
                sn_pri(xav1e->tdev, SN_ERR, "%s,vf_index incorrect in ONE_VF_MODE\n", __func__);
            } else {
                ncores = sizeof(onevf_core_array)/sizeof(onevf_core_array[0]);
                cp     = onevf_core_array;
            }
            break;

        default:
            sn_pri(xav1e->tdev, SN_ERR, "%s,incorrect pf_vf_mode (%d)\n", __func__, xav1e->tdev->pf_vf_mode);
            break;
    }

    *num_cores = ncores;
    return cp;
}


static int init_core_array(struct xav1_enc *xav1e)
{
    int i, irq;
    u32 total_io_size = 0;
    struct xav1_enc_core_desc* cd = NULL;
    int num_cores = 0;

    //get core description table for pf/vf
    cd = get_core_desc(xav1e, &num_cores);
    if (!num_cores) {
        sn_pri(xav1e->tdev, SN_ERR, "failed to get num_cores\n");
        return 0;
    }

    for (i = 0; i < num_cores; i++) {
        xav1e->subsys[i].base_addr = (cd+i)->base;
        xav1e->subsys[i].iosize    = (cd+i)->iosize;
        irq                        = (cd+i)->irq;
        if (irq != -1)
            xav1e->subsys[i].irq   = pci_irq_vector(xav1e->tdev->pdev, irq);

        total_io_size              += xav1e->subsys[i].iosize;
        xav1e->subsys[i].hwreg     = 0;
        xav1e->subsys[i].name      = (cd+i)->name;

        sn_pri(xav1e->tdev, SN_DBG, "xav1_enc:slice[%d]: base=%x  iosize=%d  tot_size=%d  core_irq id: %d  pci_irq: %d\n",
                i, xav1e->subsys[i].base_addr, xav1e->subsys[i].iosize, total_io_size,
                irq, xav1e->subsys[i].irq);
    }

    return total_io_size;
}

static int reserve_io(struct xav1_enc *xav1e)
{
    int i;

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if (xav1e->subsys[i].iosize) {
            sn_pri(xav1e->tdev, SN_DBG, "xav1_enc:slice[%d]: bar2_virt=0x%x  ipbase=0x%lx, iosize=%d\n", i,
                    xav1e->tdev->bar2_virt, xav1e->subsys[i].base_addr, xav1e->subsys[i].iosize);

            xav1e->subsys[i].hwreg = xav1e->tdev->bar2_virt + xav1e->subsys[i].base_addr;

            if (xav1e->subsys[i].hwreg == NULL) {
                sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: failed to ioremap HW %d regs\n", i);
                return -EBUSY;
            }
            sn_pri(xav1e->tdev, SN_DBG, "xav1_enc:slice[%d]: hwreg(bar2_virt+base)=0x%x  reg[0]=0x%08x\n",
                i, xav1e->subsys[i].hwreg, readl(xav1e->subsys[i].hwreg));
        }
    }
    return 0;
}

static int check_slice(struct xav1_enc *xav1e, unsigned slice) {
    if (slice >= MAX_SLICE_NUM || !xav1e->subsys[slice].base_addr || !xav1e->slices[slice].valid) {
        TRACE("Invalid slice: %u\n", slice);
        return -1;
    }
    return 0;
}

static int send_msg_raw(struct xav1_enc *Enc, struct xav1_msg *Msg, int Lock)
{
    volatile u8* hwregs;
    int i;
    unsigned long flags;
    hwregs = Enc->subsys[Msg->slice].hwreg;
    if (Lock) {
        spin_lock_irqsave(&Enc->reserve_lock, flags);
    }
    {
        int rawVal = readl(hwregs + XAV1_MBOX0_PTR);
        int writeIdx = (rawVal >> 8) & 127;
        int readIdx  = (rawVal >> 20) & 127;
        int avail = ((readIdx + 128 - writeIdx) & 127) + 128 * !!(rawVal & 1);
        if (avail < Msg->count) {
            if (Lock) {
                spin_unlock_irqrestore(&Enc->reserve_lock, flags);
            }
            TRACE("No room for message: %08x %d %d %d %d\n", rawVal, writeIdx, readIdx, avail, Msg->count);
            ++Enc->mb_full[Msg->slice];
            return -EAGAIN;
        }
    }
    for (i = 0; i < Msg->count - 1; ++i) {
        writel(Msg->content[i], hwregs + XAV1_MBOX0_DATA);
    }
    writel(Msg->content[i], hwregs + XAV1_MBOX0_LAST);
    if (Lock) {
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    }
    return 0;
}

void patch_ring(struct xav1_enc* Enc, struct file* Fp, unsigned Slice, unsigned Slot, uint32_t Write, uint32_t Watermark, unsigned Dir) {
    struct slice* slice = Enc->slices + Slice;
    struct slot* slot = slice->slots + Slot;
    uint8_t* ring = (Dir) ? slot->h2fw_ring : slot->fw2h_ring;
    uint8_t* data = ring + 2 * sizeof(uint32_t);
    uint32_t lastWrite = (Dir) ? slot->h2fw_write : slot->fw2h_write;
    TRACE("SL%u RING%u: %016llx  W: %u [%u]  K: %u  R: %u\n", Slot, Dir, (uint64_t) ring, Write, lastWrite, Watermark, *((uint32_t*) ring));
    if (!Dir) {
        while (lastWrite != Write) {
            if  (lastWrite == Watermark) {
                lastWrite = 0;
            }
            switch (data[lastWrite]) {
            case 0: { // FW2HOST_HW_BITSTREAM
                uint64_t* patch = (uint64_t*) (data + lastWrite + 16);
                *patch = sn_mem_osal_translate_mem(Enc->tdev, slot->task_id, *patch);
                lastWrite += 32;
                TRACE("*FW2HOST_HW_BITSTREAM\n");
                break;
            }
            case 1: // FW2HOST_FW_BITSTREAM
                lastWrite += 20 + *(uint32_t*) (data + lastWrite + 8);
                TRACE("*FW2HOST_FW_BITSTREAM\n");
                break;
            case 2: // FW2HOST_METADATA_AV1
            case 3: { // FW2HOST_METADATA_VCE
                uint64_t* patch = (uint64_t*) (data + lastWrite + 8);
                if (patch[0] != 1) {
                    patch[0] = sn_mem_osal_translate_mem(Enc->tdev, slot->task_id, patch[0]);
                }
                if (patch[1] != 1) {
                    patch[1] = sn_mem_osal_translate_mem(Enc->tdev, slot->task_id, patch[1]);
                }
                lastWrite += (data[lastWrite] == 2) ? 160 : 120;
                TRACE("*FW2HOST_METADATA_*\n");
                break; }
            case 4: // FW2HOST_OP_STATS
                lastWrite += 24;
                TRACE("*FW2HOST_OP_STATS\n");
                break;
            default: {
                TRACE("***** UNKNOWN FATAL *****\n");
                lastWrite = Write;
                break; }
            }
            TRACE("WRITE: %u of %u\n", lastWrite, Write);
        }
        slot->fw2h_write = lastWrite;
        return;
    }
    while (lastWrite != Write) {
        if  (lastWrite == Watermark) {
            lastWrite = 0;
        }
        switch (data[lastWrite]) {
        case 0: { // HOST2FW_VIDEOFRAME
            uint64_t* patch = (uint64_t*) &data[lastWrite] + 1;
            int i;
            for (i = 0; i < 7; ++i) {
                if (patch[i] != 1) { // INVALID_ADDR
                    patch[i] = sn_mem_osal_translate_handle(Enc->tdev, Fp, patch[i], 0);
                }
            }
            lastWrite += 112;
            TRACE("*HOST2FW_VIDEOFRAME\n");
            break; }
        case 1: // HOST2FW_DESCRIPTOR_STATUS
            TRACE("*HOST2FW_DESCRIPTOR_STATUS\n");
            lastWrite += 12;
            break;
        case 2: // HOST2FW_DESCRIPTOR_VUIPARAMS
            TRACE("*HOST2FW_DESCRIPTOR_VUIPARAMS\n");
            lastWrite += 12;
            break;
        default:
            TRACE("***** UNKNOWN FATAL *****\n");
            lastWrite = Write;
            break;
        }
    }
    slot->h2fw_write = lastWrite;
}

static int ddr_patchin_check(struct xav1_enc* Enc, struct xav1_msg* Msg) { // Only patchable msg is MAILBOX_MSG_FW2HOST_RING_NOTIFY
    TRACE("RING_FW2HOST\n");
    patch_ring(Enc, 0, Msg->slice, (Msg->content[0] >> 10) & 255, Msg->content[1] & 0xffff, Msg->content[1] >> 16, 0);
    return 0;
}

static int process_host_messages(struct xav1_enc *Enc, int Slice)
{
    // TODO(jkuhn) error handling and recovery
    volatile u8 *hwregs = Enc->subsys[Slice].hwreg;
    struct fifo_msg msg;
    u32 header;
    int numWords;
    int slotNum;
    int arg;
    msg.words[0] = readl(hwregs + XAV1_MBOX15_DATA);
    header = msg.words[0];
    numWords = (header >> 8) & 3;
    if (numWords) {
        int i;
        for (i = 1; i < numWords + 1; ++i) {
            msg.words[i] = readl(hwregs + XAV1_MBOX15_DATA);
        }
    }
    if ((header & 255) == 252) {
        int cpu = header >> 30;
        if (numWords) {
            int index = (header >> 10) & 0xfffff;
            int i;
            for (i = 1; i < numWords + 1; ++i) {
                printk(KERN_CRIT "[BT] %08x %d Device %s Slice %d Cpu %d\n", msg.words[i], index + i - 1, Enc->tdev->dev_name, Slice, cpu);
            }
        } else {
            int error = (header >> 10) & 0xfffff;
            Enc->fatal_fw_error = true;
            printk(KERN_CRIT "FATAL FPS ERROR: %08x Device %s Slice %d Cpu %d\n", error, Enc->tdev->dev_name, Slice, cpu);
        }
        return 0;
    }
    if ((header & 255) == 253) {
        u32 ipId = (header >> 11) & 127;
        u32 load = (header >> 18);
        if (ipId >= SN_PERF_ID_COUNT) {
          TRACE("INGORING UNKNOWN IP: %u", ipId);
          return 0;
        }
        sn_perf_load(Enc->perfHandles[ipId], Slice, load);
        return 0;
    }
    if ((header & 255) == 254) {
        sn_perf_event event;
        event.bits = header >> 10;
        event.hwTs = msg.words[1];
        event.payload = msg.words[2];
        event.name = msg.words[3];
        // TRACE("%u.%03u.%02u %u %9u %08x %.4s\n", event.slice, event.ipId, event.unitId, event.type, event.hwTsValid, event.hwTs, event.namec);
        if (event.ipId >= SN_PERF_ID_COUNT) {
          TRACE("INGORING UNKNOWN IP: %u", event.ipId);
          return 0;
        }
        sn_perf_record(Enc->perfHandles[event.ipId], &event);
        return 0;
    }
    slotNum = (header >> 10) & 0xff;
    arg = (header >> 18);
    // TRACE("Got msg %d for %d + %d [%x]\n", header & 255, slotNum, numWords, header);
    if (slotNum < MAX_SLOTS_PER_SLICE) {
        struct slot* slot = Enc->slices[Slice].slots + slotNum;
        if (!slot->filp) {
            TRACE("DROPPING message for unallocated slot %u:%u\n", Slice, slotNum);
            return 0;
        }
        if (slot->pendingDelete) {
            if ((header & 255) == 255) { // teardown
                void* filp;
                struct slice* slice = Enc->slices + Slice;
                unsigned long flags;
                spin_lock_irqsave(&Enc->reserve_lock, flags);
                filp = slot->filp;
                slot->filp = NULL;
                ++slice->numChan;
                spin_unlock_irqrestore(&Enc->reserve_lock, flags);
                TRACE("Freed slot: %u:%u [%u] for %p\n", Slice, slotNum, slice->numChan, filp);
                return 0;
            }
            TRACE("Ignoring message for slot %u:%u pending delete\n", Slice, slotNum);
            return 0;
        }
        if (!kfifo_put(&slot->fifo, msg)) {
            slot->fatal_error = true;
            sn_pri(Enc->tdev, SN_ERR, "**FATAL** Full slot %u:%u\n", Slice, slotNum);
        } else {
            TRACE("SL%u MSG\n", slotNum);
#if defined(USE_EVENTS)
            if (slot->eventFile) {
                struct eventfd_ctx* ctx = eventfd_ctx_fileget(slot->eventFile);
                if (ctx) {
                    TRACE("SIGNALLING\n");
                    eventfd_signal(ctx, 1);
                    eventfd_ctx_put(ctx);
                }
            }
#endif
            wake_up_all(&slot->wait_queue);
        }
    }
    return 0;
}

static irqreturn_t isr(int irq, void *dev_id)
{
    unsigned long flags;
    volatile u8 *hwregs;
    u32 irq_status;
    int slice = 0;
    struct xav1_enc *xav1e = (struct xav1_enc *) dev_id;
    u32  int_shadow_status_off = S1_INT_SD_XENC_STUS;
    u32 int_shadow_isr_off = S1_INT_SD_XENC_ISR;

    // TRACE("ISR\n");
    spin_lock_irqsave(&xav1e->irq_lock, flags);
    //determine slice index from irq#
    if(irq == xav1e->subsys[0].irq) {
        slice = 0;
        int_shadow_status_off = S1_INT_SD_XENC_STUS;
        int_shadow_isr_off = S1_INT_SD_XENC_ISR;
    } else if(irq == xav1e->subsys[1].irq) {
        slice = 1;
        int_shadow_status_off = S2_INT_SD_XENC_STUS;
        int_shadow_isr_off = S2_INT_SD_XENC_ISR;
    } else {
        sn_pri(xav1e->tdev, SN_ERR, "xav1: error:: irq %d not supported \n", irq);
    }

    hwregs = xav1e->subsys[slice].hwreg;

    /* Semaphore HOST interrupt status register read */
    irq_status = readl(hwregs + XAV1_SAR_HOST_INTR_STATUS);
    //check if Host IRQ
    // TRACE("irq_status (shadow): %08x read: %08x\n", irq_status, readl(hwregs + XAV1_SAR_HOST_INTR_STATUS));
    if (irq_status & XBIT_MASK(29)) {
        process_host_messages(xav1e, slice);
    }

    spin_unlock_irqrestore(&xav1e->irq_lock, flags);

    return IRQ_HANDLED;
}

#if SUB_SYS_VCE != 1 /* assume legoVSI FPGA which doesn't have xav1 HW */
static void internal_configure_on(struct xav1_enc *xav1e)
{
    int i;
    u32 val;
    volatile u8 *base_addr=NULL;
    unsigned int loop;

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        base_addr =  xav1e->tdev->bar2_virt + ((xav1e->tdev->pf_vf_mode == PF_MODE) ? pf_core_array[i].base : onevf_core_array[i].base);
		loop = 1000;
        val = readl(base_addr + XAV1_SAR_CONTROL0_CLR);
        // writel(val|0x1000,base_addr + THS0_POWER_CTRL_SET);
        // printk("%s %d, set reg base=0x%x, poll reg base=0x%x\n",__func__,__LINE__,(base_addr+THS0_POWER_CTRL_SET),(base_addr+THS0_POWER_CTRL_POLL));
        writel(0x1000,base_addr + XAV1_SAR_CONTROL0_CLR);

        val = readl(base_addr + XAV1_SAR_FPS_STATUS);
        while ((val&0x3000000) != 0x0) {
            val = readl(base_addr + XAV1_SAR_FPS_STATUS);
            msleep(1);
            loop--;
            if (!loop) {
                sn_pri(xav1e->tdev,SN_ERR,"xav1 internal power on configuration error!\n");
                break;
            }
        }
    }
}

static void internal_configure_off(struct xav1_enc *xav1e)
{
    int i;
    u32 val;
    volatile u8 *base_addr=NULL;
    unsigned int loop;

    for (i = 0; i < MAX_SLICE_NUM; i++) {
		loop = 1000;
        base_addr =  xav1e->tdev->bar2_virt + ((xav1e->tdev->pf_vf_mode == PF_MODE) ? pf_core_array[i].base : onevf_core_array[i].base);
        val = readl(base_addr + XAV1_SAR_CONTROL0_SET);
        writel(0x1000,base_addr + XAV1_SAR_CONTROL0_SET);

        val = readl(base_addr + XAV1_SAR_FPS_STATUS);
        while ((val&0x3000000) != 0x3000000) {
            val = readl(base_addr + XAV1_SAR_FPS_STATUS);
            msleep(1);
            loop--;
            if (!loop) {
                sn_pri(xav1e->tdev,SN_ERR,"xav1 internal power off  configuration error!\n");
                break;
            }
        }
    }
}
#endif

// reclaim all allocated slots and prevent new allocations
static int cancel_slot_allocation(struct xav1_enc* Enc, int SliceNo)
{
    unsigned long flags;
    if (SliceNo < 0 || SliceNo >= MAX_SLICE_NUM) {
        return -EINVAL;
    }
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    {
        struct slice* slice = Enc->slices + SliceNo;
        struct slot* slot = slice->slots;
        int j;
        slice->numChan = 0;
        for (j = 0; j < MAX_SLOTS_PER_SLICE; ++j, ++slot) {
            slot->filp = NULL;
        }
    }
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    return 0;
}

// allow allocations again after cancel_slot_allocation
static int begin_slot_allocation(struct xav1_enc* Enc, int SliceNo)
{
    unsigned long flags;
    if (SliceNo < 0 || SliceNo >= MAX_SLICE_NUM) {
        return -EINVAL;
    }
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    Enc->slices[SliceNo].numChan = MAX_SLOTS_PER_SLICE;
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    return 0;
 }

int xav1_whole_subsys_soft_reset(struct xav1_enc *xav1e, int slice)
{
    volatile u8 *hwregs = NULL;
    if (slice < 0 || slice >= MAX_SLICE_NUM)
    {
        return -1;
    }
    hwregs = xav1e->subsys[slice].hwreg;

    sys_xav1_whole_subsys_soft_reset(xav1e->tdev, slice);
    cancel_slot_allocation(xav1e, slice);
    sys_xav1_whole_subsys_release_soft_reset(xav1e->tdev, slice);
    writel(1 << 29, hwregs + XAV1_SAR_CONTROL0_SET); // enable MBOX15 host interrupt
    begin_slot_allocation(xav1e, slice);
    return 0;
}

static int init_slices(struct xav1_enc* Enc)
{
    struct slice* slice = Enc->slices;
    int i;
    TRACE("init_slices\n");
    for (i = 0; i < MAX_SLICE_NUM; ++i, ++slice) {
        struct slot* slot = slice->slots;
        volatile u8 *hwregs = Enc->subsys[i].hwreg;
        int j;
        slice->numChan = MAX_SLOTS_PER_SLICE;
        slice->nextAvailSlot = 1;
        for (j = 0; j < MAX_SLOTS_PER_SLICE; ++j, ++slot) {
            slot->filp = NULL;
        }
        slice->valid = readl(hwregs + XAV1_MBOX0_PTR) != 0; // TODO(jkuhn): robust?
        writel(1 << 29, hwregs + XAV1_SAR_CONTROL0_SET); // enable MBOX15 host interrupt
        writel(0, hwregs + XAV1_MBOX15_PTR); // reset mailboxes
        writel(0, hwregs + XAV1_MBOX0_PTR); // reset mailboxes
        mutex_init(&slice->lock);
    }
    return 0;
}

static int xav1_enc_perf(struct sn_tranx_t* tdev, __u32 ipId, __u32 cmd, __u32 arg) {
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    struct xav1_msg msg;
    struct slice* slice = xav1e->slices;
    int i;
    for (i = 0; i < MAX_SLICE_NUM; ++i, ++slice) {
        if (!slice->valid) {
            continue;
        }
        msg.slice = i;
        msg.count = 1;
        msg.content[0] = 254 | (ipId << 10) | (cmd << 17) | (arg << 19);
        send_msg_raw(xav1e, &msg, 1); // TODO(jkuhn): spin???
        // TRACE("%u %u %u %u\n", i, ipId, cmd, arg);
    }
    return 0;
}

static int alloc_slot(struct xav1_enc* Enc, struct file* fp, osal_command* cmd, osal_response* resp) {
    struct slice* slice;
    struct slot* slot;
    struct slot* end;
    struct xav1_slot slotInfo;
    unsigned long flags;
    signed int ret;

    slotInfo.slice = cmd->data[0];

    if (check_slice(Enc, slotInfo.slice)) {
        ret = -EINVAL;
        goto exit;
    }
    slice = Enc->slices + slotInfo.slice;
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    if (!slice->numChan) {
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
        ret = -ENOMEM;
        goto exit;
    }
    slot = slice->slots + slice->nextAvailSlot;
    end = slot;
    slotInfo.slot = slice->nextAvailSlot;
    do {
        if (!slot->filp) {
            slot->filp = fp;
            slot->pendingDelete = 0;
            slot->fatal_error = false;
            slice->nextAvailSlot = (slotInfo.slot + 1) % MAX_SLOTS_PER_SLICE;
            --slice->numChan;
            spin_unlock_irqrestore(&Enc->reserve_lock, flags);
            init_waitqueue_head(&slot->wait_queue);
            spin_lock_init(&slot->fifo_lock);
            INIT_KFIFO(slot->fifo);
            TRACE("Allocated slot: %u:%u [%u] for %p\n", slotInfo.slice, slotInfo.slot, slice->numChan, fp);
            resp->data[0] = slotInfo.slot;
            resp->header = cmd->header;
            resp->header.num_data_words = 1;
            return 0;
        }
        if (++slotInfo.slot == MAX_SLOTS_PER_SLICE) {
            slotInfo.slot = 0;
            slot = slice->slots;
        } else {
            ++slot;
        }
    } while (slot != end);
    // should not be possible
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    printk(KERN_CRIT "No free slots (should be)\n");
    ret = -EINVAL;

exit:
    resp->header                = cmd->header;
    resp->header.num_data_words = 1;
    resp->data[0]               = ret;
    return ret;

}

static void free_slot(struct xav1_enc *Enc, struct file *filp, osal_command* cmd, osal_response* resp) {
    struct xav1_slot slotInfo;
    struct slice* slice;
    struct slot* slot;
    unsigned long flags;
    struct xav1_msg msg;
    int ret;

    slotInfo.slice = cmd->data[0];
    slotInfo.slot = cmd->data[1];

    if (slotInfo.slot >= MAX_SLOTS_PER_SLICE || check_slice(Enc, slotInfo.slice)) {
        ret = -EINVAL;
        goto exit;
    }
    slice = Enc->slices + slotInfo.slice;
    slot = slice->slots + slotInfo.slot;
    msg.slice = slotInfo.slice;
    msg.count = 1;
    msg.content[0] = 255 | (slotInfo.slot << 10);
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    if (slot->filp != filp || slot->pendingDelete) {
        // A reset may have revoked ownership of this slot, or maybe we're dealing with an ill-behaved application.
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
        TRACE("Skipping %p %u\n", slot->filp, slot->pendingDelete);
        ret = -EINVAL;
        goto exit;
    }
    if (send_msg_raw(Enc, &msg, 0) < 0) { // no lock: We're already holding reserve_lock to exclude the possibility of reset.
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
        ret = -EAGAIN;
        goto exit;
    }
    resp->data[0]               = 0;
    resp->header                = cmd->header;
    resp->header.num_data_words = 1;
    slot->pendingDelete         = 1;
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    return;
exit:
    resp->header                = cmd->header;
    resp->header.num_data_words = 1;
    resp->data[0]               = ret;
}

static int ddr_patchout_check(struct xav1_enc* Enc, struct file* Fp, struct xav1_msg* Msg) {
    switch (Msg->content[0] & 255) { // message type;
    case 128: { // MAILBOX_MSG_PROPOSE_CHANNEL
        uint64_t raw = (((uint64_t) Msg->content[2])) << 32 | Msg->content[1];
        uint64_t* addr = (uint64_t*) sn_mem_osal_translate_mmio(Enc->tdev, raw);
        *addr = sn_mem_osal_translate_handle(Enc->tdev, Fp, *addr, 0);
        TRACE("MSG_PROPOSE: %016llx\n", *addr);
        return 0;
    }
    case 129: { // MAILBOX_MSG_COMMIT_CHANNEL
        uint64_t raw = (((uint64_t) Msg->content[2])) << 32 | Msg->content[1];
        uint32_t* addr = (uint32_t*) sn_mem_osal_translate_mmio(Enc->tdev, raw);
        uint32_t count = addr[CHANNEL_CONFIG_SIZE_WORDS];
        struct ResourceInst {
            uint8_t  m_type;
            uint32_t m_index;
            uint64_t m_addr;
        }* inst = (struct ResourceInst*) (addr + CHANNEL_CONFIG_SIZE_WORDS + 1);
        int i;
        TRACE("MSG_COMMIT\n");
        TRACE("Patching %d resources\n", count);
        for (i = 0; i < count; ++i, ++inst) {
            inst->m_addr = sn_mem_osal_translate_handle(Enc->tdev, Fp, inst->m_addr, 0);
            TRACE("%3u %u %016llx\n", i, inst->m_type, inst->m_addr);
        }
        return 0;
    }
    case 131: { // MAILBOX_MSG_HOST2FW_RING_NOTIFY
        TRACE("SL%u MSG_HOST2FW\n", (Msg->content[0] >> 10) & 255);
        patch_ring(Enc, Fp, Msg->slice, (Msg->content[0] >> 10) & 255, Msg->content[1] & 0xffff, Msg->content[1] >> 16, 1);
        return 0;
    }
    default:
        break;
    }
    return 0;
}

static int send_msg(struct xav1_enc* Enc, struct file* Fp, osal_command* cmd, osal_response* resp) {
    struct xav1_msg msg;
    int ret;
    msg.slice = cmd->data[0];
    msg.count = cmd->data[1];
    msg.content[0] = cmd->data[2];
    msg.content[1] = cmd->data[3];
    msg.content[2] = cmd->data[4];
    msg.content[3] = cmd->data[5];
    TRACE("%08x %08x %08x %08x\n", msg.content[0], msg.content[1], msg.content[2], msg.content[3]);
    if (check_slice(Enc, msg.slice) || msg.count > 4) {
        TRACE("Invalid Message: slice: %d count: %d\n", msg.slice, msg.count);
        ret = -EINVAL;
    } else {
        struct slice* slice = Enc->slices + msg.slice;
        if (ddr_patchout_check(Enc, Fp, &msg)) {
            return 0;
        }
        mutex_lock(&slice->lock);
        ret = send_msg_raw(Enc, &msg, 1);
        mutex_unlock(&slice->lock);
    }

    resp->header                = cmd->header;
    resp->header.num_data_words = 1;
    resp->data[0]               = ret;
    return ret;
}

static int wait_msg(struct xav1_enc* Enc, struct file* fp, osal_command* cmd, osal_response* resp) {
    struct xav1_wait_msg msg;
    struct slice* slice;
    struct slot* slot;
    int ret;
    int ret_time;

    msg.slice = cmd->data[0];
    msg.slot = cmd->data[1];
    msg.timeout_ms = cmd->data[2];

    if (check_slice(Enc, msg.slice)) {
        ret = -EINVAL;
        goto exit;
    }
    slice = Enc->slices + msg.slice;
    slot = slice->slots + msg.slot;
    if (slot->filp != fp) {;
        ret = -EINVAL;
        goto exit;
    }
    if (slot->fatal_error) {
        ret = -ENOTRECOVERABLE;
        goto exit;
    }
    ret_time = wait_event_interruptible_timeout(slot->wait_queue, !kfifo_is_empty(&slot->fifo), msg.timeout_ms * HZ / 1000);
    TRACE("SL%u WAITED: %d %d\n", msg.slot, msg.timeout_ms * HZ / 1000, ret_time);
    if (ret_time < 0) {
        ret = -EFAULT;
        goto exit;
    }
    resp->data[0] = ret_time * 1000 / HZ;
    resp->header.num_data_words = 1;
    if (ret_time) {
        int i;
        struct fifo_msg val;
        if (!kfifo_get(&slot->fifo, &val)) {
            TRACE("FIFO was empty!\n");
            ret = -EINVAL;
            goto exit;
        }
        for (i = 0; i < MAX_WORDS_PER_MSG; ++i) {
            resp->data[i + 1] = val.words[i];
        }
        if ((val.words[0] & 255) == 132) { // MAILBOX_MSG_FW2HOST_RING_NOTIFY
            struct xav1_msg msg2;
            msg2.slice = msg.slice;
            msg2.content[0] = val.words[0];
            msg2.content[1] = val.words[1];
            msg2.content[2] = val.words[2];
            msg2.content[3] = val.words[3];
            ddr_patchin_check(Enc, &msg2);
        }
        resp->header.num_data_words = 5;
    }
    resp->header                = cmd->header;
    return 0;

exit:
    resp->header                = cmd->header;
    resp->header.num_data_words = 1;
    resp->data[0]               = ret;
    return ret;
}

static int set_ring_addrs(struct xav1_enc* Enc, struct file* Fp, osal_command* cmd, osal_response* resp) {
    struct slice* slice;
    struct slot* slot;
    if (check_slice(Enc, cmd->data[0])) {
        return -EINVAL;
    }
    slice = Enc->slices + cmd->data[0];
    slot = slice->slots + cmd->data[1];
    if (slot->filp != Fp || slot->pendingDelete) {
        TRACE("Skipping %p %u\n", slot->filp, slot->pendingDelete);
        return -EINVAL;
    }
    slot->h2fw_ring = sn_mem_osal_translate_mmio(Enc->tdev, *((uint64_t*) (cmd->data + 2)));
    slot->h2fw_write = 0;
    slot->fw2h_ring = sn_mem_osal_translate_mmio(Enc->tdev, *((uint64_t*) (cmd->data + 4)));
    slot->fw2h_write = 0;
    slot->task_id = sn_mem_osal_task_from_handle(*((uint64_t*) (cmd->data + 6)));
    TRACE("SL%u TASK: %u  H2FW: %016llx  FW2H: %016llx\n", cmd->data[1], slot->task_id, (uint64_t) slot->h2fw_ring, (uint64_t) slot->fw2h_ring);
    return 0;
}

#if defined(USE_EVENTS)

static int associate_event(struct xav1_enc* Enc, struct task_struct* Task, struct file* Fp, osal_command* cmd, osal_response* resp) {
    struct file* file;
    struct slice* slice;
    struct slot* slot;
    if (check_slice(Enc, cmd->data[0])) {
        return -EINVAL;
    }
    slice = Enc->slices + cmd->data[0];
    slot = slice->slots + cmd->data[1];
    if (slot->filp != Fp || slot->pendingDelete) {
        TRACE("Skipping %p %u\n", slot->filp, slot->pendingDelete);
        return -EINVAL;
    }
    rcu_read_lock();
    file = files_lookup_fd_raw(Task->files, cmd->data[2]);
    rcu_read_unlock();
    if (!file) {
        TRACE("Unable to find eventfd: %u\n", cmd->data[2]);
        return -EINVAL;
    }
    slot->eventFile = file;
    TRACE("Set eventFile for %u:%u\n", cmd->data[0], cmd->data[1]);
    resp->header                = cmd->header;
    resp->header.num_data_words = 1;
    resp->data[0]               = 0;
    return 0;
}

#endif

static void osal_encoder_handler(struct work_struct *workPtr)
{
    sn_osal_work* work       = (sn_osal_work*) workPtr;
    osal_command* command    = work->cmdList.cmds;
    osal_response* response  = work->rspList.rsps;
    struct sn_tranx_t* tdev  = work->tdev;
    struct xav1_enc* xav1e   = tdev->modules[SN_MODULE_XAV1_ENC];
    struct file* fp          = work->filp;
#if defined(USE_EVENTS)
    struct task_struct* task = work->task;
#endif
    int i                    = 0;
    long ret                 = 0;

    work->rspList.num_rsps = work->cmdList.num_cmds;

    //command list respresents transaction and could encapsulate multiple commands
    for (i = 0; i <  work->cmdList.num_cmds; ++i, ++command, ++response)
    {
        switch(command->header.op_code)
        {
        case OSAL_ACCEL_CMD_ENC_ALLOC_SLOT:
            TRACE("ALLOC_SLOT\n");
            ret = alloc_slot(xav1e, fp, command, response);
            break;
        case OSAL_ACCEL_CMD_ENC_FREE_SLOT:
            TRACE("FREE_SLOT\n");
            free_slot(xav1e, fp, command, response);
            break;
        case OSAL_ACCEL_CMD_ENC_SEND_MSG:
            TRACE("SEND_MSG\n");
            ret = send_msg(xav1e, fp, command, response);
            break;
        case OSAL_ACCEL_CMD_ENC_WAIT_MSG:
            TRACE("WAIT_MSG\n");
            ret = wait_msg(xav1e, fp, command, response);
            break;
        case OSAL_ACCEL_CMD_ENC_SET_RING_ADDRS:
            TRACE("RING_ADDRS\n");
            ret = set_ring_addrs(xav1e, fp, command, response);
            break;
#if defined(USE_EVENTS)
        case OSAL_ACCEL_CMD_GLOBAL_ASSOCIATE_EVENT:
            TRACE("ASSOCIATE_EVENT\n");
            ret = associate_event(xav1e, task, fp, command, response);
            break;
#endif
        default:
            sn_pri(tdev, SN_ERR, "xav1_enc: %s, OSAL cmd:0x%x invalid.\n", __func__, command->header.op_code);
            response->header                = command->header;
            response->header.num_data_words = 1;
            response->data[0]               = 0xdeadbeef;
            break;
        }
    }
    sn_osal_finish_work(work);
}

static ssize_t enable_load_store(struct device* Dev, struct device_attribute* Attr, const char* Buf, size_t Count)
{
    struct sn_misc_tdev* mtdev = dev_get_drvdata(Dev);
    struct sn_tranx_t* tdev = mtdev->tdev;
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    struct xav1_msg msg;
    if (Count != 2 || (Buf[0] != '0' && Buf[0] != '1')) {
        sn_pri(tdev, SN_ERR, "Must be 0 or 1 only\n");
        return -EINVAL;
    }
    msg.slice = 0; // send to both slices;
    msg.count = 1;
    msg.content[0] = 251 | ((Buf[0] - '0') << 10);
    send_msg_raw(xav1e, &msg, 1);
    msg.slice = 1;
    send_msg_raw(xav1e, &msg, 1);
    return Count;
}

static ssize_t mb_full_show(struct device* Dev, struct device_attribute* Attr, char* Buf)
{
    struct sn_misc_tdev* mtdev = dev_get_drvdata(Dev);
    struct sn_tranx_t* tdev = mtdev->tdev;
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    return sprintf(Buf, "%u %u\n", xav1e->mb_full[0], xav1e->mb_full[1]);
}

static ssize_t fps_ecc_counts_show(struct device* Dev, struct device_attribute* Attr, char* Buf)
{
    struct sn_misc_tdev* mtdev = dev_get_drvdata(Dev);
    struct sn_tranx_t* tdev = mtdev->tdev;
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    int len = 0, slice, cpu;
    for (slice = 0; slice < 2; ++slice) {
        for (cpu = 0; cpu < 8; ++cpu) {
            len += sprintf(Buf + len, "S%uC%u%c  CO: %10u  UC: %10u\n", slice, cpu & 3, (cpu < 4) ? 'I' : 'D', xav1e->ecc_counts[slice][cpu][0], xav1e->ecc_counts[slice][cpu][1]);
        }
    }
    return len;
}

static DEVICE_ATTR_WO(enable_load);
static DEVICE_ATTR_RO(mb_full);
static DEVICE_ATTR_RO(fps_ecc_counts);

static struct attribute* trans_xav1_sysfs_entries[] = {
    &dev_attr_enable_load.attr, &dev_attr_mb_full.attr, &dev_attr_fps_ecc_counts.attr, NULL
};

static struct attribute_group trans_xav1_attribute_group = {
    .name = NULL,
    .attrs = trans_xav1_sysfs_entries,
};

int xav1_get_register(struct sn_tranx_t *tdev, u32 slice_num, u32 reg)
{
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    return readl(xav1e->subsys[slice_num].hwreg + reg);
}

int xav1_soft_reset(struct sn_tranx_t *tdev)
{
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];

    if (xav1_whole_subsys_soft_reset(xav1e, 0) != 0) {
        return -1;
    }

    if (xav1_whole_subsys_soft_reset(xav1e, 1) !=0) {
        return -1;
    }

    return 0;
}

static void ecc_counts_timer(struct timer_list* timer)
{
    struct xav1_enc* xav1e = from_timer(xav1e, timer, ecc_counts_timer);
    int slice, cpu;
    for (slice = 0; slice < 2; ++slice) {
        volatile u8* hwregs = xav1e->subsys[slice].hwreg;
        for (cpu = 0; cpu < 8; ++cpu) {
            int rawVal = readl(hwregs + XAV1_SAR_ECC_COUNTS + cpu * 4);
            xav1e->ecc_counts[slice][cpu][0] = rawVal & ((1 << 20) - 1);
            xav1e->ecc_counts[slice][cpu][1] = rawVal >> 20;
        }
    }
    mod_timer(timer, timer->expires + 5 * HZ); // 5 seconds
}

int xav1_enc_init(struct sn_tranx_t *tdev)
{
    int i, result;
    struct xav1_enc *xav1e;
    int total_iosize;
    void *tmp_mem;

    xav1e = (struct xav1_enc *) kvzalloc(sizeof(*xav1e), GFP_KERNEL);
    if (!xav1e) {
        sn_pri(tdev, SN_ERR, "xav1_enc: kmalloc failed\n");
        return -ENOMEM;
    }
    tdev->modules[SN_MODULE_XAV1_ENC] = xav1e;
    xav1e->tdev = tdev;

    total_iosize = init_core_array(xav1e);
    if (!total_iosize) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: total core count is zero\n");
        goto err_free_xav1e;
    }

#if SUB_SYS_VCE != 1 /* assume legoVSI FPGA which doesn't have xav1 HW */
    internal_configure_on(xav1e);
#endif

    spin_lock_init(&xav1e->irq_lock);
    spin_lock_init(&xav1e->reserve_lock);

    result = sysfs_create_group(&tdev->misc_dev->this_device->kobj, &trans_xav1_attribute_group);
    if (result) {
        sn_pri(tdev, SN_ERR, "xav1: failed to create sysfs device attributes\n");
        goto err_free_xav1e;
    }

    result = reserve_io(xav1e);
    if(result < 0) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: reserve io failed\n");
        goto err_free_xav1e;
    }

    tmp_mem = vzalloc(total_iosize);
    if (!tmp_mem) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: malloc regs mem failed.\n");
        goto err_free_xav1e;
    }

    xav1e->regs_shadow = tmp_mem;
    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if (xav1e->subsys[i].iosize) {
            xav1e->subsys[i].shadow = tmp_mem;
            tmp_mem += xav1e->subsys[i].iosize;
        }
    }

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if(xav1e->subsys[i].irq > 0) {
            result = request_irq(xav1e->subsys[i].irq, isr,
                                 IRQF_SHARED, xav1e->subsys[i].name, (void *)xav1e);
            if(result != 0) {
                if(result == -EINVAL) {
                    sn_pri(tdev, SN_ERR,
                        "xav1_enc: Bad IRQ:%d or handler\n",
                        xav1e->subsys[i].irq);
                } else if(result == -EBUSY) {
                    sn_pri(tdev, SN_ERR,
                        "xav1_enc: IRQ:%d busy\n",
                        xav1e->subsys[i].irq);
                }
                goto err_free_irq;
            } else {
                sn_pri(tdev, SN_INF, "xav1_enc:slice[%d]: pci_irq %d registered!\n", i, xav1e->subsys[i].irq);
            }
        } else
            sn_pri(tdev, SN_INF, "xav1_enc: IRQ:%d not in use!\n", i);
    }
    init_slices(xav1e);
    xav1e->perfHandles[SN_PERF_ID_VPP] = sn_perf_register(tdev, "enc_vpp", SN_PERF_ID_VPP, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_ME] = sn_perf_register(tdev, "enc_me", SN_PERF_ID_ME, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_VCE] = sn_perf_register(tdev, "enc_vce", SN_PERF_ID_VCE, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_AV1] = sn_perf_register(tdev, "enc_av1", SN_PERF_ID_AV1, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_SYSTEMCPU] = sn_perf_register(tdev, "enc_systemcpu", SN_PERF_ID_SYSTEMCPU, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_INLINECPU] = sn_perf_register(tdev, "enc_inlinecpu", SN_PERF_ID_INLINECPU, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_LEGOCPU0] = sn_perf_register(tdev, "enc_legocpu0", SN_PERF_ID_LEGOCPU0, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_LEGOCPU1] = sn_perf_register(tdev, "enc_legocpu1", SN_PERF_ID_LEGOCPU1, xav1_enc_perf, 2);
    xav1e->perfHandles[SN_PERF_ID_CUTREE] = sn_perf_register(tdev, "enc_cutree", SN_PERF_ID_CUTREE, xav1_enc_perf, 2);

    sema_init(&xav1e->boot_sem, 1);

    /* load encoder firmware to FPS sub system */
    result = load_fps_subsystem(tdev);
    if(result) {
        goto err_free_irq;
    }

    xav1e->fatal_fw_error = false;

    timer_setup(&xav1e->ecc_counts_timer, ecc_counts_timer, 0);
    mod_timer(&xav1e->ecc_counts_timer, jiffies);
    sn_osal_register(tdev, "encoder", osal_encoder, osal_encoder_handler, NULL);
    sn_pri(tdev, SN_INF, "encoder: osal handler registered\n");

    sn_pri(tdev, SN_INF, "xav1_enc: module initialization done\n");
    return 0;

err_free_irq:
    for (i= 0; i < MAX_SLICE_NUM; i++) {
        if(xav1e->subsys[i].irq > 0)
          free_irq(xav1e->subsys[i].irq, (void *)xav1e);
    }
    vfree(xav1e->regs_shadow);

err_free_xav1e:
    kfree(xav1e);
    sn_pri(tdev, SN_ERR, "xav1_enc: module not inserted\n");
    return result;
}

void xav1_enc_release(struct sn_tranx_t *tdev)
{
    struct xav1_enc *xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    int i;

    del_timer_sync(&xav1e->ecc_counts_timer);
    /* free the IRQ */
    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if(xav1e->subsys[i].irq > 0) {
            xav1_whole_subsys_soft_reset(xav1e, i);
            free_irq(xav1e->subsys[i].irq, (void *)xav1e);
        }
    }
#if SUB_SYS_VCE != 1
    internal_configure_off(xav1e);
#endif

    vfree(xav1e->regs_shadow);

    sysfs_remove_group(&tdev->misc_dev->this_device->kobj, &trans_xav1_attribute_group);

    sn_pri(tdev, SN_DBG, "xav1_enc: remove module done.\n");
}

long xav1_enc_ioctl(struct file *filp,
                    unsigned int cmd,
                    unsigned long argp,
                    struct sn_tranx_t *tdev)
{
    long ret = 0;
    struct xav1_enc *xav1e = tdev->modules[SN_MODULE_XAV1_ENC];

    if (xav1e->fatal_fw_error && cmd != XAV1_ENC_IOCTL_TWO_SLICE_SOFT_RESET) {
        return -ENOTRECOVERABLE;
    }

    switch (cmd) {
    case XAV1_ENC_IOCTL_TWO_SLICE_SOFT_RESET:
        ret = xav1_whole_subsys_soft_reset(xav1e, 0);
        ret |= xav1_whole_subsys_soft_reset(xav1e, 1);
        break;
    default:
        sn_pri(tdev, SN_ERR,
            "xav1_enc: %s, cmd:0x%x error.\n", __func__, cmd);
        ret = -EINVAL;
    }

    /* After XAV1 reset, load encoder firmwares to FPS sub system */
    if (cmd == XAV1_ENC_IOCTL_TWO_SLICE_SOFT_RESET) {
        ret = load_fps_subsystem(tdev);
        xav1e->fatal_fw_error = false;
    }

    return ret;
}

void xav1_close(struct sn_tranx_t *tdev, struct file *filp)
{
    struct xav1_enc *xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    struct slice* slice = xav1e->slices;
    int i;
    TRACE("xav1_close\n");
    for (i = 0; i < MAX_SLICE_NUM; ++i, ++slice) {
        unsigned long flags;
        int j;
        struct slot* slot = slice->slots;
        spin_lock_irqsave(&xav1e->reserve_lock, flags);
        for (j = 0; j < MAX_SLOTS_PER_SLICE; ++j, ++slot) {
            if (slot->filp == filp && !slot->pendingDelete) {
                int retry = 1000;
                slot->pendingDelete = 1;
                {
                    struct xav1_msg msg;
                    msg.slice = i;
                    msg.count = 1;
                    msg.content[0] = 255 | (j << 10);
                    while (retry-- && send_msg_raw(xav1e, &msg, 0) < 0) { // no lock: We're already holding reserve_lock to exclude the possibility of reset.
                        spin_unlock_irqrestore(&xav1e->reserve_lock, flags);
                        msleep(1);
                        spin_lock_irqsave(&xav1e->reserve_lock, flags);
                    }
                }
                spin_unlock_irqrestore(&xav1e->reserve_lock, flags);
                TRACE("Closing dangling slot %d:%d for %p\n", i, j, filp);
                if (retry < 0) {
                    xav1e->fatal_fw_error = true;
                    sn_pri(tdev, SN_ERR, "Firmware is unresponsive on slice %d\n", slice - xav1e->slices);
                }
                spin_lock_irqsave(&xav1e->reserve_lock, flags);
            }
        }
        spin_unlock_irqrestore(&xav1e->reserve_lock, flags);
    }

    if (filp == xav1e->boot_sem_owner) {
        xav1e->boot_sem_owner = NULL;
        TRACE("%p CLOSE RELEASE\n", filp);
        up(&xav1e->boot_sem);
    }
}
